package com.wudsn.productions.atari2600.sv2k17.scroll;

import java.awt.Color;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import com.wudsn.productions.atari2600.sv2k17.AbstractGenerator;

/**
 * SV2K17 - Sine Scroll Generator
 * 
 * Generates and visualizes all DYCP combinations and checks for collisions.
 * Optimizes vertical event lanes and generated 2-3 banks of code. Optimizes run
 * lengths of consecutive sequences via JSR chains and JSR/JMP replacements.
 * Originally coded in 2013 for Silly Venture 2k14.
 * 
 * @author Peter Dell/JAC!
 */
public final class ScrollGenerator extends AbstractGenerator {

    public static void main(String[] args) {

	main(ScrollGenerator.class, args);
    }

    private static final String MAIN_SOURCE_FILE_ = "../../SV2K17.asm";
    private static final String OUTPUT_FILE_PATH = "C:\\jac\\system\\Atari2600\\Programming\\Demos\\SV2K17\\scroll\\frames\\SV2K17-Scroll-Frame-{0}.asm";

    private static final int POSITIONING_HEIGHT = 1;
    private static final int CHARACTER_HEIGHT = 7;
    private static final int CHARACTER_HEIGHT_PADDING = 1;
    private static final int MIN_N = 6;
    private static final int MAX_N = 160 / 8;
    private static final int NUM_N = MAX_N - MIN_N + 1;

    private static final int X_FRAME_WIDTH = 160;
    private static final int Y_FRAME_HEIGHT = 192;
    private static final int X_INT_WIDTH = 8;
    private static final int XPOS_OFFSET = 0x30;

    private static final class EventType {
	public static final int NONE = 0;
	public static final int POSITION = 1;
	public static final int DRAW = 2;
	public static final int CLEAR = 3;
    }

    private static final class Event {
	public static final Event NONE_P0 = new Event(0, EventType.NONE, -1, -1);
	public static final Event NONE_P1 = new Event(1, EventType.NONE, -1, -1);

	public final int player;
	public final int type;
	public final int character_index;
	public final int line_index;

	public Event(int player, int type, int character_index, int line_index) {
	    this.player = player;
	    this.type = type;
	    this.character_index = character_index;
	    this.line_index = line_index;
	}

	@Override
	public String toString() {
	    switch (type) {
	    case EventType.NONE:
		return "P" + player + " off";
	    case EventType.POSITION:
		return "P" + player + " positioning character " + character_index;
	    case EventType.DRAW:
		return "P" + player + " drawing character " + character_index + " line " + line_index;
	    case EventType.CLEAR:
		return "P" + player + " clearing";
	    default:
		throw new RuntimeException();
	    }
	}
    }

    private final static class Lane {
	public boolean value;
	public boolean oldValue;
	public int count;

	public Lane() {
	    value = false;
	    oldValue = false;
	    count = 0;
	}

	public void setValue(boolean value) {
	    this.value = value;
	    if (value == true) {
		count++;
	    }
	}

	public boolean isEnding() {
	    return oldValue && !value;
	}

	public void nextLine() {
	    oldValue = value;
	}
    }

    private static final class Players {
	public int[] xpos_tab;
	public List<List<Event>> lineEvents; // First index is line, second
					     // index is player

	public Players(int visible_height, int n) {
	    xpos_tab = new int[n];
	    lineEvents = new ArrayList<List<Event>>(visible_height);
	    for (int i = 0; i < visible_height; i++) {
		List<Event> playerEvents = new ArrayList<Event>(2);
		playerEvents.add(Event.NONE_P0);
		playerEvents.add(Event.NONE_P1);
		lineEvents.add(playerEvents);
	    }
	}

	public int size() {
	    int result = 2 * lineEvents.size();
	    return result;
	}

	public Event getEvent(int line, int player) {
	    return lineEvents.get(line).get(player);
	}

	public void setEvent(int line, int player, int type, int character_index, int line_index) {
	    Event event = getEvent(line, player);
	    if (event.type != EventType.NONE) {
		if (event.type == EventType.CLEAR || type == EventType.POSITION) {
		    // In lines where RESP is strobed, the Sprite is not
		    // displayed
		} else {
		    throw new RuntimeException("Event already present: " + event.toString());
		}
	    }
	    lineEvents.get(line).set(player, new Event(player, type, character_index, line_index));
	}
    }

    private static final class Parameters implements Cloneable {
	public Parameters() {
	    player_frames = new ArrayList<Players>();
	}

	public int wave_steps;
	public int positioning_height;
	public int character_height;
	public int amplitude;
	public int n;
	public int collisions;
	public List<Players> player_frames;
	public int[] positioning_counts;

	public int getEffectiveCharacterHeight() {
	    return positioning_height + character_height + CHARACTER_HEIGHT_PADDING;
	}

	public int getVisibleFrameHeight() {
	    return amplitude + getEffectiveCharacterHeight();
	}

	@Override
	public String toString() {
	    return "character_height=" + character_height + " positioning_height=" + positioning_height
		    + " wave_steps=" + wave_steps + " amplitude=" + amplitude + " n=" + n + " collisions=" + collisions
		    + " player_frames=" + player_frames.size() + " player_frames_size=" + getSize(player_frames)
		    + " bytes";
	}

	public Parameters copy() {
	    Parameters result = new Parameters();
	    result.wave_steps = wave_steps;
	    result.character_height = character_height;
	    result.positioning_height = positioning_height;
	    result.amplitude = amplitude;
	    result.player_frames = new ArrayList<Players>(player_frames);
	    result.n = n;
	    result.collisions = collisions;
	    return result;
	}

	public int getSize(List<Players> player_frames) {
	    if (player_frames == null) {
		throw new IllegalArgumentException("Parameter 'player_frames' must not be null.");
	    }
	    int result = 0;
	    for (Players players : player_frames) {
		result += players.size();
	    }
	    return result;
	}
    }

    private static final class ParametersComparator implements Comparator<Parameters> {

	public ParametersComparator() {

	}

	@Override
	public int compare(Parameters arg0, Parameters arg1) {
	    int result = 0;
	    if (result == 0) {
		result = -(arg0.n - arg1.n);
	    }
	    if (result == 0) {
		result = arg0.character_height - arg1.character_height;
	    }
	    if (result == 0) {
		result = arg0.wave_steps - arg1.wave_steps;
	    }
	    return result;
	}
    }

    private final static class Block {
	public Block(int i, int xpos_int, int ypos_int, int ypos_end_int) {
	    this.i = i;
	    this.xpos_int = xpos_int;
	    this.ypos_int = ypos_int;
	    this.ypos_end_int = ypos_end_int;
	}

	public final int i;
	public final int xpos_int;
	public final int ypos_int;
	public final int ypos_end_int;

	@Override
	public String toString() {
	    return "i=" + i;
	}
    }

    private final class BlockComparator implements Comparator<Block> {
	public BlockComparator() {

	}

	@Override
	public int compare(Block o1, Block o2) {
	    return o1.ypos_int - o2.ypos_int;
	}

    }

    @Override
    protected void run() {
	for (int i = 0; i < 255; i++) {

	}
	int width = X_FRAME_WIDTH * (X_INT_WIDTH * 3);
	int height = Y_FRAME_HEIGHT * NUM_N;
	createGraphics(width, height);
	List<Parameters> parametersList;

	parametersList = new ArrayList<Parameters>();
	Parameters parameters;
	parameters = new Parameters();

	parameters.positioning_height = POSITIONING_HEIGHT;
	parameters.character_height = CHARACTER_HEIGHT;
	parameters.amplitude = 152;
	parameters.wave_steps = 360;
	int ypos_screen_base_int = 0;
	for (int n = MIN_N; n < MAX_N; n++) {
	    parameters.n = n;

	    ypos_screen_base_int = evaluateParameters(parameters, true, ypos_screen_base_int);
	    if (parameters.collisions == 0) {
		parametersList.add(parameters.copy());
	    }
	}

	Collections.sort(parametersList, new ParametersComparator());
	Parameters bestParameters = parametersList.get(0);
	println("Best Parameters:");
	println(bestParameters);

	evaluateParameters(bestParameters, true, ypos_screen_base_int);

	boolean printEvents = false;
	if (printEvents) {
	    printEvents(bestParameters);
	}

	adaptParameters(bestParameters);
	analyzeParameters(bestParameters);
	int frame_count = bestParameters.player_frames.size();

	for (int i = 0; i < frame_count; i++) {
	    List<String> lines = renderParameters(bestParameters, i, i+1);
	    writeParameters(bestParameters, i, lines);
	}
    }

    private int evaluateParameters(Parameters parameters, boolean draw, int ypos_screen_base_int) {
	if (draw) {
	    println("\nEvaluate Parameters:");
	}
	double amplitude_half = (parameters.amplitude - 1) / 2;
	int visible_height = parameters.getVisibleFrameHeight();
	int[] collision_array = new int[visible_height];
	final double rad_factor = Math.PI / 180;
	int n = parameters.n;
	double alpha_step = (parameters.wave_steps) / n;
	double xpos_screen_base_double = 0;
	double xpos_step_double = (X_FRAME_WIDTH - X_INT_WIDTH) / n;
	parameters.collisions = 0;
	int x_screen_frame_base_int = 0;

	parameters.player_frames.clear();
	BlockComparator blockComparator = new BlockComparator();
	for (double alpha_0 = 0.0d; alpha_0 < alpha_step; alpha_0 = alpha_0 + 1) {

	    double alpha = alpha_0;
	    List<Block> blocks = new ArrayList<Block>(n);
	    for (int i = 0; i < n; i++) {
		int xpos_int = (int) (xpos_screen_base_double + i * xpos_step_double);
		double sin = Math.sin(alpha * rad_factor);
		double ypos_double = amplitude_half * sin;
		int ypos_int = (int) (ypos_double + amplitude_half);
		int ypos_end_int = ypos_int + parameters.getEffectiveCharacterHeight();
		Block block = new Block(i, xpos_int, ypos_int, ypos_end_int);
		blocks.add(block);
		alpha += alpha_step;
	    }

	    Collections.sort(blocks, blockComparator);

	    int x_screen_base_int = x_screen_frame_base_int;
	    boolean collision = false;
	    Arrays.fill(collision_array, 0);
	    Players players = new Players(visible_height, n);
	    if (draw) {
		println("Frame " + parameters.player_frames.size() + ":" + blocks.toString());
	    }
	    parameters.player_frames.add(players);
	    int usedPlayer = 0;
	    for (int j = 0; j < n; j++) {
		Block block = blocks.get(j);
		int player_xpos = block.xpos_int + XPOS_OFFSET;
		players.xpos_tab[block.i] = player_xpos;
		int x2_int = x_screen_base_int + block.i * X_INT_WIDTH + 1;
		for (int y = block.ypos_int; y < block.ypos_end_int; y++) {
		    collision_array[y]++;
		    int collisionCount = collision_array[y];
		    Color color;

		    if (collisionCount == 1 || collisionCount == 2) {
			color = usedPlayer == 0 ? Color.BLUE : Color.GREEN;

		    } else {
			parameters.collisions++;
			collision = true;
			color = Color.RED;
		    }

		    if (draw) {
			graphics.setColor(color);
			graphics.drawLine(x2_int + 1, ypos_screen_base_int + y, x2_int + X_INT_WIDTH,
				ypos_screen_base_int + y);
		    }
		}
		if (!collision) {
		    int y = block.ypos_int;
		    int character_index = block.i;
		    if (draw) {
			println("Player " + usedPlayer + ", Character Index " + character_index + " in line "
				+ block.ypos_int);
		    }

		    players.setEvent(y++, usedPlayer, EventType.POSITION, character_index, -1);
		    for (int i = 0; i < parameters.character_height; i++) {
			players.setEvent(y++, usedPlayer, EventType.DRAW, character_index, i);

		    }
		    players.setEvent(y++, usedPlayer, EventType.CLEAR, -1, -1);
		}
		usedPlayer = 1 - usedPlayer;

	    }
	    if (draw) {
		graphics.setColor(collision ? Color.red : Color.blue);
		graphics.drawRect(x_screen_base_int, ypos_screen_base_int, n * X_INT_WIDTH + 1, Y_FRAME_HEIGHT + 1);
	    }
	    x_screen_frame_base_int += n * X_INT_WIDTH + 2;

	} // for(alpha_0...)

	if (draw) {
	    graphics.setColor(parameters.collisions > 0 ? Color.red : Color.blue);
	    graphics.drawString("char_n=" + n, 1, ypos_screen_base_int + 16);
	    graphics.drawString("frame_n=" + parameters.player_frames.size(), 1, ypos_screen_base_int + 28);
	    graphics.drawString("collisions=" + parameters.collisions, 1, ypos_screen_base_int + 40);

	}
	ypos_screen_base_int = ypos_screen_base_int + Y_FRAME_HEIGHT + 2;
	return ypos_screen_base_int;
    }

    private void adaptParameters(Parameters parameters) {
	if (parameters == null) {
	    throw new IllegalArgumentException("Parameter 'parameters' must not be null.");
	}
	int fixes = 0;
	List<Players> player_frames = parameters.player_frames;
	for (int i = 0; i < player_frames.size(); i++) {
	    Players players = player_frames.get(i);

	    // Solve double positioning
	    for (int y = 0; y < parameters.getVisibleFrameHeight(); y++) {
		Event player0_event = players.lineEvents.get(y).get(0);
		Event player1_event = players.lineEvents.get(y).get(1);

		if (player0_event.type == EventType.POSITION && player1_event.type == EventType.POSITION) {
		    println("ERROR: Frame " + i + " has double positioning in single line " + y);
		    // Try to position player0 one line earlier
		    int player = 0;
		    Event player_event_above = players.lineEvents.get(y - 1).get(player);
		    if (player_event_above.type == EventType.NONE) {
			players.lineEvents.get(y - 1).set(player, player0_event);
			players.lineEvents.get(y).set(player, player_event_above);
			println("INFO:  Solved using player " + player + ". New positioning in line " + (y - 1));
			fixes++;
		    } else {
			// Player0 didn't work, try player1
			player = 1;
			player_event_above = players.lineEvents.get(y - 1).get(player);

			if (player_event_above.type == EventType.NONE) {
			    players.lineEvents.get(y - 1).set(player, player1_event);
			    players.lineEvents.get(y).set(player, player_event_above);
			    println("INFO:  Solved using player " + player + ". New positioning in line " + (y - 1));
			    fixes++;
			} else {
			    throw new RuntimeException("ERROR: Cannot resolve double positioning: "
				    + player0_event.toString() + "/" + player0_event.toString());
			}
		    }
		}
	    }

	    // Solve close positioning
	    for (int y = 0; y < parameters.getVisibleFrameHeight() - 1; y++) {
		int player0_y = y;
		int player1_y = y + 1;
		Event player0_event = players.lineEvents.get(player0_y).get(0);
		Event player1_event = players.lineEvents.get(player1_y).get(1);

		if (player0_event.type == EventType.POSITION && player1_event.type == EventType.POSITION) {
		    println("ERROR: Frame " + i + " has close positioning P0 before P1 in line " + player0_y + "&"
			    + player1_y);
		    // Try to position player0 one line earlier
		    int player = 0;
		    Event player_event_above = players.lineEvents.get(y - 1).get(player);
		    if (player_event_above.type == EventType.NONE || player_event_above.type == EventType.CLEAR) {
			players.lineEvents.get(y - 1).set(player, player0_event);
			players.lineEvents.get(y).set(player, player_event_above);
			println("INFO: Solved using player " + player + ". New positioning in line " + (y - 1));
			fixes++;
		    } else {
			throw new RuntimeException("ERROR: Cannot resolve close positioning: "
				+ player0_event.toString() + "/" + player1_event.toString());
		    }
		}

		player0_y = y + 1;
		player1_y = y;
		player0_event = players.lineEvents.get(player0_y).get(0);
		player1_event = players.lineEvents.get(player1_y).get(1);

		if (player0_event.type == EventType.POSITION && player1_event.type == EventType.POSITION) {
		    println("ERROR: Frame " + i + " has close positioning P1 before P0 in line " + player1_y + "&"
			    + player0_y);
		    // Try to position player1 one line earlier
		    int player = 1;
		    Event player_event_above = players.lineEvents.get(y - 1).get(player);
		    if (player_event_above.type == EventType.NONE) {
			players.lineEvents.get(y - 1).set(player, player1_event);
			players.lineEvents.get(y).set(player, player_event_above);
			println("INFO:  Solved using player " + player + ". New positioning in line " + (y - 1));
			fixes++;
		    } else {
			throw new RuntimeException("ERROR: Cannot resolve close positioning: "
				+ player0_event.toString() + "/" + player0_event.toString());
		    }
		}
	    }

	}
	println("INFO:  " + fixes + " conflicts fixed automatically.");
    }

    private void analyzeParameters(Parameters parameters) {
	if (parameters == null) {
	    throw new IllegalArgumentException("Parameter 'parameters' must not be null.");
	}
	int[] position_counts = new int[parameters.getVisibleFrameHeight()];
	List<Players> player_frames = parameters.player_frames;
	for (int i = 0; i < player_frames.size(); i++) {
	    Players players = player_frames.get(i);
	    for (int y = 0; y < parameters.getVisibleFrameHeight() - 1; y++) {
		if (players.getEvent(y, 0).type == EventType.POSITION
			|| players.getEvent(y, 1).type == EventType.POSITION) {
		    position_counts[y]++;
		}
	    }
	}

	parameters.positioning_counts = position_counts;
    }

    private void printEvents(Parameters parameters) {
	println("\nEvents per Frame:");
	List<Players> player_frames = parameters.player_frames;
	int frame_count = player_frames.size();
	for (int frame = 0; frame < frame_count; frame++) {
	    Players players = player_frames.get(frame);
	    int y_max = parameters.getVisibleFrameHeight();
	    for (int y = 0; y < y_max; y++) {
		Event player0_event = players.lineEvents.get(y).get(0);
		Event player1_event = players.lineEvents.get(y).get(1);

		println("Frame " + frame + ": y=" + y + " " + player0_event.toString() + "/" + player1_event.toString());
	    }
	}

    }

    private List<String> renderParameters(Parameters parameters, int startFrame, int endFrame) {
	if (parameters == null) {
	    throw new IllegalArgumentException("Parameter 'parameters' must not be null.");
	}
	println("\nRendering Parameters from frame " + startFrame + " to " + endFrame + ":");
	List<Players> player_frames = parameters.player_frames;
//	int frame_count = player_frames.size();
//	int rendered_frame_count = 0;

	List<String> allLines = new ArrayList<String>();
//	allLines.add("\t.proc frames");
	String prefix = "\t";
	int onlyFrame = -1;
	int frame;
	for (frame = startFrame; frame < endFrame; frame++) {
	    if (onlyFrame >= 0 && onlyFrame != frame) {
		continue;
	    }
//	    rendered_frame_count++;

	    List<String> lines = new ArrayList<String>();
	    println("Frame " + frame);
	    lines.add("\t.proc frame_" + frame);
	    Players players = player_frames.get(frame);

	    boolean positioned_old = false;
	    boolean hmclr_required = false;
	    boolean hmclr_possible = false;
	    boolean hmclr_already = false;

	    Lane wsync_required = new Lane();
	    Lane draw_p0_and_p1_with_sync = new Lane();
	    Lane draw_p0_with_sync = new Lane();
	    Lane draw_p1_with_sync = new Lane();
	    Lane clear_p0_and_p1_with_sync = new Lane();

	    int y_max = parameters.getVisibleFrameHeight();
	    for (int y = 0; y < y_max; y++) {
		Event player0_event = players.lineEvents.get(y).get(0);
		Event player1_event = players.lineEvents.get(y).get(1);

		if (player0_event.type == EventType.POSITION && player1_event.type == EventType.POSITION) {
		    throw new RuntimeException("ERROR: Double positioning in line " + y + ": "
			    + player0_event.toString() + "/" + player1_event.toString());
		}

		String lineComment = "\t;Line " + y + ", " + player0_event.toString() + ", " + player1_event.toString();

		boolean positioned;
		positioned = false;
		if (player0_event.type == EventType.POSITION) {
		    println("P0 character_index=" + player0_event.character_index + " in line " + y);
		    lines.add("\nsprite_" + to2DigitNumer(player0_event.character_index) + "_p0" + lineComment);

		    lines.add(prefix + "m_lda_xpos_char " + (player0_event.player) + " "
			    + (player0_event.character_index));

		    // if (player0_event.character_index == 0) {
		    // lines.add(prefix + "lda xpos_cnt");
		    // } else {
		    // lines.add(prefix + "lda xpos_tab+" +
		    // (player0_event.character_index - 1));
		    // }

		    if (player1_event.type == EventType.DRAW || player1_event.type == EventType.CLEAR) {
			if (player1_event.type == EventType.DRAW) {
			    lines.add(prefix + "ldx char_tab+" + player1_event.character_index + "*char_height+"
				    + player1_event.line_index);
			} else {
			    lines.add(prefix + "ldx #0");
			}
			lines.add(prefix + "mx_set_hpos p0 grp1");
			player1_event = Event.NONE_P1;
		    } else {
			lines.add(prefix + "mx_set_hpos p0");
		    }
		    positioned = true;
		} else if (player1_event.type == EventType.POSITION) {
		    println("P1 character_index=" + player1_event.character_index + " in line " + y);
		    lines.add("\nsprite_" + to2DigitNumer(player1_event.character_index) + "_p1" + lineComment);

		    lines.add(prefix + "m_lda_xpos_char " + (player1_event.player) + " "
			    + (player1_event.character_index));

		    // if (player1_event.character_index == 0) {
		    // lines.add(prefix + "lda xpos_cnt");
		    // } else {
		    // lines.add(prefix + "lda xpos_tab+" +
		    // (player1_event.character_index - 1));
		    // }
		    if (player0_event.type == EventType.DRAW || player0_event.type == EventType.CLEAR) {
			if (player0_event.type == EventType.DRAW) {
			    lines.add(prefix + "ldx char_tab+" + player0_event.character_index + "*char_height+"
				    + player0_event.line_index);
			} else {
			    lines.add(prefix + "ldx #0");
			}
			lines.add(prefix + "mx_set_hpos p1 grp0");
			player0_event = Event.NONE_P0;

		    } else {
			lines.add(prefix + "mx_set_hpos p1");
		    }
		    positioned = true;
		}

		if (positioned) {
		    hmclr_required = true;
		    hmclr_possible = false;
		    hmclr_already = false;
		}

		// Compute player graphics update mode.
		// Draw for a single player also always clears the other
		// player.
		// Clear for a single player also always clears the other
		// player.
		draw_p0_and_p1_with_sync.setValue(player0_event.type == EventType.DRAW
			&& player1_event.type == EventType.DRAW);
		draw_p0_with_sync.setValue(player0_event.type == EventType.DRAW
			&& (player1_event.type != EventType.DRAW));
		draw_p1_with_sync.setValue(player1_event.type == EventType.DRAW
			&& (player0_event.type != EventType.DRAW));
		clear_p0_and_p1_with_sync.setValue(player0_event.type == EventType.CLEAR
			&& (player1_event.type != EventType.DRAW) || player1_event.type == EventType.CLEAR
			&& (player0_event.type != EventType.DRAW));

		boolean draw_or_clear_with_sync = draw_p0_and_p1_with_sync.value || draw_p0_with_sync.value
			|| draw_p1_with_sync.value || clear_p0_and_p1_with_sync.value;

		boolean wsync_will_be_provided = draw_or_clear_with_sync;

		boolean wsync_already = positioned || positioned_old;
		wsync_required.setValue(!wsync_already && !wsync_will_be_provided);

		boolean wsync_flush_required = (y == y_max - 1)
			|| wsync_required.isEnding()
			|| wsync_will_be_provided
			|| (player0_event.type == EventType.DRAW || player0_event.type == EventType.CLEAR
				|| player1_event.type == EventType.DRAW || player1_event.type == EventType.CLEAR || hmclr_required);

		if (wsync_flush_required) {
		    final int WSYNC_COUNT_JSR_MIN = 2;
		    final int WSYNC_COUNT_JSR_MAX = 16;

		    // STA WSYNC is shorter than JSR WSYNC1
		    while (wsync_required.count > 0) {
			if (wsync_required.count < WSYNC_COUNT_JSR_MIN) {
			    for (int w = 0; w < wsync_required.count; w++) {
				lines.add(prefix + "sta wsync;" + (w + 1) + "/" + wsync_required.count);
			    }
			    wsync_required.count = 0;
			} else {
			    int count = Math.min(wsync_required.count, WSYNC_COUNT_JSR_MAX);
			    lines.add(prefix + "jsr wait_wsync.r" + count);
			    wsync_required.count -= count;
			    // if (hmclr_required) {
			    // hmclr_already = true;
			    // }
			}
		    }
		}

		if (draw_p0_and_p1_with_sync.value) {
		    // Load new indexes only if required
		    if (!draw_p0_and_p1_with_sync.oldValue) {
			// if (wsync_already) {
			// lines.add(prefix + "ldx char_tab+" +
			// player0_event.character_index + "*char_height+"
			// + player0_event.line_index);
			// lines.add(prefix + "ldy char_tab+" +
			// player1_event.character_index + "*char_height+"
			// + player1_event.line_index);
			// lines.add(prefix +
			// "jsr object.store_grp0_gr1_from_xy");
			// hmclr_already = true;
			// }
			lines.add(prefix + "ldx #" + player0_event.character_index + "*char_height+"
				+ player0_event.line_index);
			lines.add(prefix + "ldy #" + player1_event.character_index + "*char_height+"
				+ player1_event.line_index);
		    }
		    if (wsync_already) {
			// Note: This causes player1 to be not here if it is too
			// far left
			lines.add(prefix + "jsr object.store_grp0_grp1.without_sync");
		    } else {
			lines.add(prefix + "jsr object.store_grp0_grp1.with_sync");
		    }
		} else if (clear_p0_and_p1_with_sync.value) {
		    if (wsync_already) {
			lines.add(prefix + "jsr object.clear_grp0_grp1.without_sync");
		    } else {
			lines.add(prefix + "jsr object.clear_grp0_grp1.with_sync");
		    }
		    // hmclr_already = true;
		} else {
		    if (player0_event.type == EventType.DRAW) {

			if (draw_p0_with_sync.value) {
			    // Load new index only if required
			    if (!draw_p0_with_sync.oldValue && !draw_p0_and_p1_with_sync.oldValue) {
				lines.add(prefix + "ldx #" + player0_event.character_index + "*char_height+"
					+ player0_event.line_index);
			    }
			    if (wsync_already) {
				lines.add(prefix + "jsr object.store_grp0.without_sync");
			    } else {
				lines.add(prefix + "jsr object.store_grp0.with_sync");
			    }
			} else {
			    throw new RuntimeException("Should never happen");
			    // lines.add(prefix + "lda char_tab+" +
			    // player0_event.character_index + "*char_height+"
			    // + player0_event.line_index);
			    // lines.add(prefix + "sta grp0");

			}
			// hmclr_already = true;
		    }

		    if (player1_event.type == EventType.DRAW) {

			if (draw_p1_with_sync.value) {
			    // Load new index only if required
			    if (!draw_p1_with_sync.oldValue && !draw_p0_and_p1_with_sync.oldValue) {
				lines.add(prefix + "ldy #" + player1_event.character_index + "*char_height+"
					+ player1_event.line_index);
			    }
			    if (wsync_already) {
				lines.add(prefix + "jsr object.store_grp1.without_sync");
			    } else {
				lines.add(prefix + "jsr object.store_grp1.with_sync");
			    }
			} else {
			    throw new RuntimeException("Should never happen");
			    // lines.add(prefix + "lda char_tab+" +
			    // player1_event.character_index + "*char_height+"
			    // + player1_event.line_index);
			    // lines.add(prefix + "sta grp1");
			}
			// hmclr_already = true;
		    }
		}

		if (hmclr_required) {
		    if (!hmclr_possible) {
			hmclr_possible = true;
		    } else {
			if (!hmclr_already) {
			    if (!draw_or_clear_with_sync) {
				lines.add(prefix + "jsr object.wait_and_hmclr");
			    } else {
				lines.add(prefix + "sta hmclr");
			    }
			}

			hmclr_required = false;
		    }
		}

		if (parameters.positioning_counts[y] == 0) {
		    // lines.add("// No positioning in this line");
		    // lines.add(prefix + "lda #" + y);
		    // lines.add(prefix + "sta colubk");
		}
		positioned_old = positioned;

		draw_p0_and_p1_with_sync.nextLine();
		draw_p0_with_sync.nextLine();
		draw_p1_with_sync.nextLine();
		clear_p0_and_p1_with_sync.nextLine();

		wsync_required.nextLine();

	    }

	    lines.add("\tm_end_frame");
	    lines.add("\t.endp");
	    lines.add("\tm_info frame_" + frame);

	    lines.add("");

	    compactLines(lines);

	    allLines.addAll(lines);
	}
//	allLines.add("\t.endp");
//	allLines.add("\tm_info frames");
//	allLines.add("\t.print \"Estimated size \", [[.len frames]*" + frame_count + "]/" + rendered_frame_count);
	return allLines;

    }

    private void compactLines(List<String> lines) {
	if (lines == null) {
	    throw new IllegalArgumentException("Parameter 'lines' must not be null.");
	}
	String oldLine = lines.get(0);
	int oldLineIndex = 0;
	int oldLineCount = 0;
	for (int i = 1; i < lines.size(); i++) {

	    String line = lines.get(i);
	    if (line.equals(oldLine)) {
		if (line.contains("jsr")) {
		    if (oldLineCount == 0) {
			oldLineCount = 1;
		    }
		    oldLineCount++;
		}
	    } else {
		if (oldLineCount > 0) {
		    for (int j = 0; j < oldLineCount; j++) {
			lines.remove(oldLineIndex);
			i--;
		    }

		    String suffix = "_r" + oldLineCount + ";Repeated";

		    if (oldLine.contains("jsr object.store_grp0_grp1.with_sync")) {
			lines.add(oldLineIndex, "\tjsr object.store_grp0_grp1" + suffix);
		    } else if (oldLine.contains("jsr object.store_grp0.with_sync")) {
			lines.add(oldLineIndex, "\tjsr object.store_grp0" + suffix);
		    } else if (oldLine.contains("jsr object.store_grp1.with_sync")) {
			lines.add(oldLineIndex, "\tjsr object.store_grp1" + suffix);
		    } else {
			throw new RuntimeException("Unhandled repetition: " + oldLine);
		    }
		    i++;
		} else if (oldLine.contains("jsr") && line.contains("rts")) {
		    lines.remove(oldLineIndex);
		    line = oldLine.replace("jsr", "jmp");
		    lines.set(oldLineIndex, line);
		    i--;
		}
		oldLine = line;
		oldLineIndex = i;
		oldLineCount = 0;
	    }
	}
    }

    private String to2DigitNumer(int x) {
	String result = String.valueOf(x);
	return x < 10 ? "0" + result : result;
    }

    private void writeParameters(Parameters parameters, int bank, List<String> lines) {
	if (parameters == null) {
	    throw new IllegalArgumentException("Parameter 'parameters' must not be null.");
	}
	if (lines == null) {
	    throw new IllegalArgumentException("Parameter 'lines' must not be null.");
	}
	// Write results
	PrintStream output;
	try {
	    String filePath = OUTPUT_FILE_PATH.replace("{0}", String.valueOf(bank));
	    output = new PrintStream(new File(filePath));
	    output.println(";\tParameters: " + parameters);
	    output.println(";");
	    output.println(";\t@com.wudsn.ide.asm.hardware=ATARI2600");
	    output.println(";\t@com.wudsn.ide.asm.mainsourcefile=" + MAIN_SOURCE_FILE_);
	    output.println(";");
	} catch (FileNotFoundException ex) {
	    println(ex.getMessage());
	    return;
	}
	try {
	    for (String line : lines) {
		output.println(line);
	    }
	} finally {
	    output.close();
	}

    }

}
